const Lang = imports.lang;
const Signals = imports.signals;

const Atk = imports.gi.Atk;
const Shell = imports.gi.Shell;
const St = imports.gi.St;
const GLib = imports.gi.GLib
const Gio = imports.gi.Gio;
const Pango = imports.gi.Pango;

const Main = imports.ui.main;

const Overview = imports.ui.overview;
const ExtensionSystem = imports.ui.extensionSystem;
const PopupMenu = imports.ui.popupMenu;
const URLHighlighter = imports.ui.messageTray.URLHighlighter;

const Config = imports.misc.config;
const ExtensionUtils  = imports.misc.extensionUtils;
const Params = imports.misc.params;

const Extension = ExtensionUtils.getCurrentExtension();

const Utils = Extension.imports.utils;

//-----------------------------------------------------------------------------------------

function launchEgo(path) {
	launchUrl(ExtensionSystem.REPOSITORY_URL_BASE + path);
}

function launchUrl(url) {
	Gio.app_info_launch_default_for_uri(url, global.create_app_launch_context());

	Main.overview.hide();
}

//-----------------------------------------------------------------------------------------

const ExtensionItem = new Lang.Class({
	Name: 'ExtensionItem',

	_init: function(params) {
		params = Params.parse(params, {
			extension: null,
			id: null
		});

		if(!params.extension || !params.id) {
			throw new Error("Must provide extension and extension ID.");
		}

		this.id = params.id;
		this.extension = params.extension;

		this.title = this.extension.metadata.name;
		this.description = this.extension.metadata.description;

		this.state = this.extension.state;

		this.error = this.state == 3; // ERROR
		this.enabled = this.state == 1; // ENABLED
		this.outOfDate = this.state == 4; // OUT_OF_DATE

		this._errorUrl = null;
		
		this._textBox = new St.BoxLayout({
			vertical: true,
		});

		this._titleBox = new St.BoxLayout({
			style_class: 'extension-title-box'
		})

		this._icon = new St.Icon({
			icon_name: 'plugin',
			style_class: 'extension-icon',
		});

		this._label =  new St.Label({
			text: GLib.markup_escape_text(this.title, -1),
			style_class: 'extension-title'
		});

		this._authorLink = new St.Button({
			label: 'Loading...',
			style_class: 'extension-author',
			x_align: St.Align.START
		});

		this._description = new URLHighlighter(this._wrap(this.description), false, false);
		this._description.actor.add_style_class_name('extension-desc');

	/*	this._description = new St.Label({
			text: GLib.markup_escape_text(this._wrap(this.description), -1),
			style_class: 'extension-desc'
		});*/

		this._websiteLink = new St.Button({
			label: 'View Website',
			style_class: 'extension-link',
			x_align: St.Align.START
		});

		this._egoLink = new St.Button({
			label: 'View Extension',
			style_class: 'extension-link',
			x_align: St.Align.START
		});
		
		this._linkBox = new St.BoxLayout({
			style_class: 'extension-links',
		})

		this._websiteLink.connect('button-release-event', 
				Lang.bind(this, this._websiteLinkClicked));
		
		this._egoLink.connect('button-release-event',
				Lang.bind(this, this._egoLinkClicked));

		this._authorLink.connect('button-release-event',
				Lang.bind(this, this._authorLinkClicked));
		
		this._textBox.add(this._titleBox);
		this._textBox.add(this._description.actor, { expand: true });
		if(this.extension.metadata.url)
			this._linkBox.add(this._websiteLink);

		this._textBox.add(this._linkBox);

		this._switch = new PopupMenu.Switch(this.enabled);

		this._switch.actor.reactive = true;
		this._switch.actor.add_style_class_name('extension-toggle');

		
		this._update = new St.Button({
			label: 'Update',
			style_class: 'extension-update'
		});

		this._update.connect('button-release-event', 
				Lang.bind(this, function() {
					// Will this work? Nobody knows!
					ExtensionSystem.uninstallExtensionFromUUID(this.id);
					ExtensionSystem.installExtensionFromUUID(this.id, '');
				})
		);

		this._error = new St.Button({
			label: 'Error',
			style_class: 'extension-error'
		});


		this._error.connect('button-release-event',
				Lang.bind(this, function() {
					launchEgo('/errors/report/' + this._pk)
				}));
		
		this._update.hide();
		this._error.hide();

		this.refresh();

		this._titleBox.add(this._update);
		this._titleBox.add(this._error);
		
		this._titleBox.add(this._switch.actor);
		this._titleBox.add(this._label);
		this._titleBox.add(this._authorLink);

		if(this.id != Extension.metadata.id) {
			this._switch.actor.connect('button-release-event', 
					Lang.bind(this, this._onToggleSwitch));
		}

		this.actor = new St.BoxLayout({
			accessible_role: Atk.Role.LIST_ITEM,
			style_class: 'extension'
		});

		this.actor.add(this._icon);
		this.actor.add(this._textBox, { expand: true });

		this.actor.queue_relayout();

		Utils.lookup(this.id, 
				Lang.bind(this, this._lookupSuccess),
				Lang.bind(this, this._lookupFailed)
			);

	},

	destroy: function() {
		// Nothing to do.
	},

	_onToggleSwitch: function(actor, event) {
		if(this.enabled) {
			log('Disabling ' + this.id);
			let exts = global.settings.get_strv(ExtensionSystem.ENABLED_EXTENSIONS_KEY);
			
			let idx = exts.indexOf(this.id);
			if(idx != -1)
				exts.splice(idx, 1);

			global.settings.set_strv(ExtensionSystem.ENABLED_EXTENSIONS_KEY, exts);
			Utils.sendPopularity(this.id, 'disable');
		} else {
			log('Enabling ' + this.id);
			let exts = global.settings.get_strv(ExtensionSystem.ENABLED_EXTENSIONS_KEY);
			if(exts.indexOf(this.id) == -1)
				exts.push(this.id);
			global.settings.set_strv(ExtensionSystem.ENABLED_EXTENSIONS_KEY, exts);
			Utils.sendPopularity(this.id, 'enable');
		}

	},

	_lookupSuccess: function(data) {

		this.data = data;
		
		this._linkBox.add(this._egoLink);
	

		let [maj, min, pt, _] = Config.PACKAGE_VERSION.split('.');
		
		let version = maj + '.' + min + '.' + pt;

		let latest = 0;

		if(data.shell_version_map[version])
			latest = data.shell_version_map[version].version;


		if(latest == 0) {
			version = maj + '.' + min;
			if(data.shell_version_map[version])
				latest = data.shell_version_map[version].version;
		}

		this._pk = data.pk;

		if(latest > this.extension.metadata.version) {
			this.outOfDate = true;
		}

		this.latest = latest;

		this._authorLink.label = GLib.markup_escape_text(data['creator'], -1);

		this._icon.gicon = new Gio.FileIcon({
			file: new Gio.file_new_for_uri(ExtensionSystem.REPOSITORY_URL_BASE + data.icon)
		});

		this.refresh();
	},

	_lookupFailed: function() {
		this.data = null;

		this._authorLink.hide();
		this._icon.gicon = new Gio.FileIcon({
			file: Extension.dir.get_child('plugin.png')
		});
	},

	_websiteLinkClicked: function(actor, event) {
		launchUrl(this.extension.metadata.url);
	},

	_authorLinkClicked: function(actor, event) {
		if(!this.data)
			return;
		
		launchEgo(this.data.creator_url);
	},

	_egoLinkClicked: function() {
		if(!this.data)
			return;

		launchEgo(this.data.link);
	},

	_wrap: function(text) {
		let out = [];
		let lines = text.split('\n');
		let len = 0;
		let buff = [];

		for(var l in lines) {
			let words = lines[l].split(' ');
			for(var i in words) {
				let word = words[i];
				if((word.length + len) > 80) {
					out.push(buff.join(' '));
					buff = [ word ];
					len = 0;
				} else {
					len += word.length + 1;
					buff.push(word);
				}
			}
			out.push(buff.join(' '));
			buff = [];
			len = 0;
		}

		return out.join('\n');
	},

	refresh: function() {
		if(this.outOfDate) {
			// We need a new batman.
			this._update.show();
			this._error.hide();
			this._switch.actor.hide();
		} else if(this.error) {
			// Oh no...
			this._update.hide();
			this._error.show();
			this._switch.actor.hide();
		} else {
			// All is well in the world.
			this._update.hide();
			this._error.hide();
			this._switch.actor.show();
		}
		
		if(this.id == Extension.uuid) {
			// Always hide the on/off switch for this actor.
			this._switch.actor.hide();
		}

		this._switch.setToggleState(this.enabled);
	},

	stateChanged: function(newState) {
		let prev_state = this.state;
	
		this.enabled = newState == 1;
		this.error = newState == 3;

		// Handle OUT_OF_DATE, though currently it doesn't change.
		if(newState == 4) {
			this.outOfDate = true;
		}

		this.state = newState;

		this.refresh();
	}

});

//-----------------------------------------------------------------------------------------

const ExtensionList = new Lang.Class({
	Name: 'ExtensionList',
	
	_init: function() {

		this._list = new St.BoxLayout({
			vertical: true,
			reactive: true,
			accessible_role: Atk.Role.LIST
		});
		
		this.actor = new St.ScrollView({
			x_fill: false,
			y_fill: false,
			style_class: 'vfade'
		});

		//this._wid = Main.intializeDeferredWork(this.actor, Lang.bind(this, this.refresh));

		this.extensions = []
		
        this.actor.add_actor(this._list);

		this._handler = ExtensionSystem.connect('extension-loaded', 
				Lang.bind(this, this._extLoaded));

		this._otherHandler = ExtensionSystem.connect('extension-state-changed',
				Lang.bind(this, this._extStateChanged));

		this._otherOtherHandler = this.actor.connect('notify::mapped',
				Lang.bind(this, this._notifyMapped));
	},

	destroy: function() {
		this.actor.destroy();

		ExtensionSystem.disconnect(this._handler);
		ExtensionSystem.disconnect(this._otherHandler);
		// Unregister the Deferred Work.
	},

	installMore: function() {
		launchEgo('/');
	},

	_extLoaded: function(source, id) {
		this._addExtension(id, ExtensionUtils.extensions[id]);	
	},

	_extStateChanged: function(source, extension) {
		let ext = this._findExtension(extension.uuid);
		if(ext)
			ext.stateChanged(extension.state);
	},

	_findExtension: function(id) {
		for(let i in this.extensions) {
			let ext = this.extensions[i];
			if(ext.id == id)
				return ext
		}
		return null;
	},

	_addExtension: function(id, ext) {
		if(!ext || this._findExtension(id)) {
			return;
		}

		let ext_item = new ExtensionItem({
			id: id,
			extension: ext
		});

		this.extensions.push(ext_item);

		this._list.add(ext_item.actor, { expand: false, x_fill: true, y_fill: false });
		
		this.actor.queue_relayout();
	},
	
	_notifyMapped: function() {
		if(this.actor.mapped)
			this.loadData();
	},

	loadData: function() {
		this.extensions = [];
		this._list.destroy_all_children();

		let installBtn = new St.Button({
			label: 'Get Extensions!',
			style_class: 'get-extensions',
			x_fill: false,
			y_fill: false,
			x_align: St.Align.MIDDLE,
			y_align: St.Align.MIDDLE
		});
		
		installBtn.connect('button-release-event',
				Lang.bind(this, this.installMore));
			
		this._list.add(installBtn);

		for(let id in ExtensionUtils.extensions) {
			this._addExtension(id, ExtensionUtils.extensions[id]);
		}
	}
});

Signals.addSignalMethods(ExtensionList.prototype);

//-----------------------------------------------------------------------------------------

let ext_list = null;

let ext_tab = null;
let ext_index = -1;

function init(meta) {

}

function enable() {
	ext_list = new ExtensionList();

	Utils.init();

	Main.overview._viewSelector.
		addViewTab('extensions', 'Extensions', ext_list.actor, 'extensions');
	
	ext_index = Main.overview._viewSelector._tabs.length - 1;
	ext_tab = Main.overview._viewSelector._tabs[ext_index];
}

function disable() {
	Utils.uninit();

	Main.overview._viewSelector._tabBox.remove_actor(ext_tab.title);
	Main.overview._viewSelector._pageArea.remove_actor(ext_tab.page);
	Main.overview._viewSelector._tabs.splice(ext_index, ext_index);

	ext_list.destroy();
	ext_list = null;
}
